/*
 *                                                                                      	
 *                Golden Multi I/O Board Device Driver                    
 *                                                                                                                
 *					 		
 *	
 *	Driver for Golden series board device driver						
 *	Based on drivers/char/serial.c, parport_pc.c, ppdev.c and lp.c
 *	by Linus Torvalds, Theodore Ts'o.
 *										
 * 	This program is free software; you can redistribute it and/or modify		
 * 	it under the terms of the GNU General Public License as published by		
 * 	the Free Software Foundation; either version 2 of the License, or		
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details. 						
 *                                                                                      
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *          
 *
 *                                                    		Version: 1.1.0.0          	
 *                                                          Date: 2008/09/05        	
 */
#include "snx_common.h"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
MODULE_LICENSE("GPL");
#endif


extern struct golden_board			golden_board_table[GOLDEN_PCI_BOARDS_MAX];
extern struct golden_ser_port	    golden_ser_table[GOLDEN_SER_TOTAL_MAX + 1];
extern struct golden_par_port	    golden_par_table[GOLDEN_PAR_TOTAL_MAX];


char snx_ser_ic_table[SNX_SER_PORT_MAX_UART][10] = 
{
	{"UNKNOWN"},
  	{"SUN1889"},
   	{"SUN1699"},
	{"SUN1999"}
};


char snx_par_ic_table[SNX_PAR_PORT_MAX_UART][10] = 
{
	{"UNKNOWN"},
  	{"SUN1888"},
   	{"SUN1689"},
	{"SUN1999"}
};


char snx_port_remap[2][10] = 
{
 	{"NON-REMAP"},
	{"REMAP"}
};


enum	
{
// serial
	GOLDEN_BOARD_TEST = 0,
	GOLDEN_BOARD_4027A,
 	GOLDEN_BOARD_4027D,
	GOLDEN_BOARD_4037A,
 	GOLDEN_BOARD_4037D,
	GOLDEN_BOARD_4036A3V,    
	GOLDEN_BOARD_4056A,
	GOLDEN_BOARD_4056D,
	GOLDEN_BOARD_4056DW,
	GOLDEN_BOARD_4066A,
	GOLDEN_BOARD_4066R,
	GOLDEN_BOARD_8139,
	GOLDEN_BOARD_8139S,
	GOLDEN_BOARD_8159,
	GOLDEN_BOARD_8159S,
	GOLDEN_BOARD_8169,
	GOLDEN_BOARD_8169S,

// parallel, support by system parport_pc driver
/* 
	GOLDEN_BOARD_4008A,
	GOLDEN_BOARD_4018A,	
*/

// multi-I/O
	GOLDEN_BOARD_4079A,
	GOLDEN_BOARD_4089A,
	GOLDEN_BOARD_4096A,	
};


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static  struct pci_device_id	golden_pci_board_id[] = 
{
// serial
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_TEST, 		0,	0,	GOLDEN_BOARD_TEST},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4027A, 	0,	0,	GOLDEN_BOARD_4027A},	
 	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4027D, 	0,	0,	GOLDEN_BOARD_4027D},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4037A, 	0,	0,	GOLDEN_BOARD_4037A},
 	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4037D, 	0,	0,	GOLDEN_BOARD_4037D},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4036A3V,	0, 	0,	GOLDEN_BOARD_4036A3V},    
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056A,  	0, 	0,	GOLDEN_BOARD_4056A},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056D,  	0, 	0,	GOLDEN_BOARD_4056D},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056DW, 	0, 	0,	GOLDEN_BOARD_4056DW},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4066A,  	0, 	0,	GOLDEN_BOARD_4066A},	
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4066R,  	0, 	0,	GOLDEN_BOARD_4066R},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8139,   	0, 	0,	GOLDEN_BOARD_8139},
	{VENID_GOLDEN,	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8139S,  	0, 	0,	GOLDEN_BOARD_8139S},
 	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8159,   	0, 	0,	GOLDEN_BOARD_8159},
 	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8159S,  	0, 	0,	GOLDEN_BOARD_8159S},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8169,   	0, 	0,	GOLDEN_BOARD_8169},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8169S,  	0, 	0,	GOLDEN_BOARD_8169S},

// parallel, support by system parport_pc driver
/*	
	{VENID_GOLDEN,	DEVID_PARALL,	SUBVENID_GOLDEN,	SUBDEVID_4008A, 	0,	0,	GOLDEN_BOARD_4008A},
	{VENID_GOLDEN,	DEVID_PARALL,	SUBVENID_GOLDEN,	SUBDEVID_4018A, 	0,	0,	GOLDEN_BOARD_4018A},	
*/

// multi-I/O
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4079A, 	0,	0,	GOLDEN_BOARD_4079A},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4089A, 	0,	0,	GOLDEN_BOARD_4089A},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4096A, 	0,	0,	GOLDEN_BOARD_4096A},	
	{0}
};
MODULE_DEVICE_TABLE(pci, golden_pci_board_id);
#else
typedef struct
{
	unsigned short vendor;
	unsigned short device;
	unsigned short subvendor;
	unsigned short subdevice;
	unsigned short driver_data;
} golden_pciInfo;


static golden_pciInfo golden_pci_board_id[] =
{
// serial
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_TEST,		GOLDEN_BOARD_TEST},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4027A,		GOLDEN_BOARD_4027A},
 	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4027D,		GOLDEN_BOARD_4027D},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4037A,		GOLDEN_BOARD_4037A},
 	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4037D,		GOLDEN_BOARD_4037D},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4036A3V, 	GOLDEN_BOARD_4036A3V},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056A,  	GOLDEN_BOARD_4056A},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056D,  	GOLDEN_BOARD_4056D},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4056DW, 	GOLDEN_BOARD_4056DW},
	{VENID_GOLDEN,	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4066A,  	GOLDEN_BOARD_4066A},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_4066R,  	GOLDEN_BOARD_4066R},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8139,   	GOLDEN_BOARD_8139},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8139S,  	GOLDEN_BOARD_8139S},
 	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8159,   	GOLDEN_BOARD_8159},
 	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8159S,  	GOLDEN_BOARD_8159S},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8169,   	GOLDEN_BOARD_8169},
	{VENID_GOLDEN, 	DEVID_SERIAL, 	SUBVENID_GOLDEN, 	SUBDEVID_8169S,  	GOLDEN_BOARD_8169S},

// parallel, support by system parport_pc driver
/*	
	{VENID_GOLDEN,	DEVID_PARALL,	SUBVENID_GOLDEN,	SUBDEVID_4008A,		GOLDEN_BOARD_4008A},
	{VENID_GOLDEN,	DEVID_PARALL,	SUBVENID_GOLDEN,	SUBDEVID_4018A,		GOLDEN_BOARD_4018A},	
*/

// multi-I/O
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4079A,		GOLDEN_BOARD_4079A},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4089A,		GOLDEN_BOARD_4089A},
	{VENID_GOLDEN,	DEVID_SERIAL,	SUBVENID_GOLDEN,	SUBDEVID_4096A,		GOLDEN_BOARD_4096A},	
	{0}
};
#endif


struct golden_board			golden_board_table[GOLDEN_PCI_BOARDS_MAX];
struct golden_ser_port	    golden_ser_table[GOLDEN_SER_TOTAL_MAX + 1];
struct golden_par_port	    golden_par_table[GOLDEN_PAR_TOTAL_MAX];


static int snx_ser_port_total_cnt;
static int snx_par_port_total_cnt;


#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18))
static irqreturn_t golden_interrupt(int irq, void *dev_id)
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static irqreturn_t golden_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#else
static void golden_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    struct golden_ser_port *sp = NULL;
    struct golden_par_port *pp = NULL;
    struct golden_board *mb = NULL;
    int i;
    int status = 0;
    
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))    
    int handled = IRQ_NONE;
#endif 
   
    for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++) 
    {
        if (dev_id == &(golden_board_table[i])) 
        {
            mb = dev_id;
            break;
        }
    }
            
    if (i == GOLDEN_PCI_BOARDS_MAX)
    {
        status = 1;
    }
    
    if (!mb)
    {
         status = 1;
    }
    
    if (mb->board_enum <= 0)
    {
         status = 1;
    }

    if (status != 0)
    {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
        return handled;
#else
        return;    
#endif 
    }

    
	if ((mb->ser_port > 0) && (mb->SerIsrFunc != NULL))
	{
		sp = &golden_ser_table[mb->ser_port_index];
	
    	if (!sp)
    	{
        	status = 1;
    	}
	
		status = mb->SerIsrFunc(mb, sp);
	}
	
	
	if ((mb->par_port > 0) && (mb->ParIsrFunc != NULL))
	{
		pp = &golden_par_table[mb->par_port_index];
	
    	if (!pp)
    	{
        	status = 1;        
    	}	
	
		status = mb->ParIsrFunc(mb, pp);	
	}


    if (status != 0)
    {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
        return handled;
#else
        return;
#endif
    }

    
    
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))            
	handled = IRQ_HANDLED;
    return handled;
#endif
}


static int golden_pci_board_probe(void)
{
    struct golden_board *mb;
    struct pci_dev *pdev = NULL;
    struct pci_dev *pdev_array[4] = {NULL, NULL, NULL, NULL};   

    int golden_pci_board_id_cnt;
    int tablecnt;
    int boardcnt;
    int i;
    unsigned short int subdeviceID;
    int status;

	// clear and init some variable
    memset(golden_board_table, 0, GOLDEN_PCI_BOARDS_MAX * sizeof(struct golden_board));

    for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
    {
        golden_board_table[i].board_enum = -1;
        golden_board_table[i].board_number = -1;
    }

    golden_pci_board_id_cnt = (sizeof(golden_pci_board_id) / sizeof(golden_pci_board_id[0])) - 1;


	// search Golden serial and multi-I/O board
	pdev = NULL;
	tablecnt = 0;
	boardcnt = 0;
	subdeviceID = 0;
	status = 0;
	
    while (tablecnt < golden_pci_board_id_cnt) 
    {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))         
        pdev = pci_get_device(VENID_GOLDEN, DEVID_SERIAL, pdev);
#else
        pdev = pci_find_device(VENID_GOLDEN, DEVID_SERIAL, pdev);        
#endif          

        if (pdev == NULL) 
        {
           	tablecnt++;
           	continue;
		}

        if ((tablecnt > 0) && ((pdev == pdev_array[0]) || (pdev == pdev_array[1]) || (pdev == pdev_array[2]) || (pdev == pdev_array[3])))
        {
            continue;
        }
            
        pci_read_config_word(pdev, 0x2e, &subdeviceID);

        if (subdeviceID == 0)
        {
            printk("GOLDEN Error: Golden Board (bus:%d device:%d), in configuration space,\n", pdev->bus->number, PCI_SLOT(pdev->devfn));
            printk("              subdevice id isn't vaild.\n\n");
            status = -EIO;
            return status;
        }

        if (subdeviceID != golden_pci_board_id[tablecnt].subdevice)
        {
            continue;
        }
        
        if (pdev == NULL)
        {
            printk("GOLDEN Error: PCI device object is an NULL pointer !\n\n");
            status = -EIO;
            return status; 
        }
        else
        {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,3))
            pci_disable_device(pdev);
#endif
            status = pci_enable_device(pdev);
            
            if (status != 0)
            { 
                printk("GOLDEN Error: Golden Board Enable Fail !\n\n");
                status = -ENXIO;
                return status;
            }
        }

        boardcnt++;
        if (boardcnt > GOLDEN_PCI_BOARDS_MAX) 
        {
            printk("\n");
            printk("GOLDEN Error: Golden Series Driver Module Support Four Boards In Maximum !\n\n");   
            status = -ENOSPC;
            return status;
        }

        mb = &golden_board_table[boardcnt-1];

        pdev_array[boardcnt-1] = pdev;
        mb->pdev = pdev;
        mb->bus_number = pdev->bus->number;
        mb->dev_number = PCI_SLOT(pdev->devfn);
		
        mb->board_enum = (int)golden_pci_board_id[tablecnt].driver_data;
        mb->pb_info = golden_pci_board_conf[mb->board_enum];
		mb->boardFlag = mb->pb_info.boardFlag;
		
        mb->board_number = boardcnt - 1;        
    }   

	
	// search Golden parallel board
	pdev = NULL;
	tablecnt = 0;
	subdeviceID = 0;
	status = 0;

    while (tablecnt < golden_pci_board_id_cnt) 
    {

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))         
        pdev = pci_get_device(VENID_GOLDEN, DEVID_PARALL, pdev);
#else
        pdev = pci_find_device(VENID_GOLDEN, DEVID_PARALL, pdev);        
#endif          

        if (pdev == NULL) 
        {
           	tablecnt++;
           	continue;
		}


        if ((tablecnt > 0) && ((pdev == pdev_array[0]) || (pdev == pdev_array[1]) || (pdev == pdev_array[2]) || (pdev == pdev_array[3])))
        {
            continue;
        }
            
        pci_read_config_word(pdev, 0x2e, &subdeviceID);

        if (subdeviceID == 0)
        {
            printk("GOLDEN Error: Golden Board (bus:%d device:%d), in configuration space,\n", pdev->bus->number, PCI_SLOT(pdev->devfn));
            printk("              subdevice id isn't vaild.\n\n");
            status = -EIO;
            return status;
        }

        if (subdeviceID != golden_pci_board_id[tablecnt].subdevice)
        {
            continue;
        }
        
        if (pdev == NULL)
        {
            printk("GOLDEN Error: PCI device object is an NULL pointer !\n\n");
            status = -EIO;
            return status; 
        }
        else
        {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,3))
            pci_disable_device(pdev);
#endif
            status = pci_enable_device(pdev);
            
            if (status != 0)
            { 
                printk("GOLDEN Error: Golden Board Enable Fail !\n\n");
                status = -ENXIO;
                return status;
            }
        }

        boardcnt++;
        if (boardcnt > GOLDEN_PCI_BOARDS_MAX) 
        {
            printk("\n");
            printk("GOLDEN Error: Golden Series Driver Module Support Four Boards In Maximum !\n\n");   
            status = -ENOSPC;
            return status;
        }

        mb = &golden_board_table[boardcnt-1];

        pdev_array[boardcnt-1] = pdev;
        mb->pdev = pdev;
        mb->bus_number = pdev->bus->number;
        mb->dev_number = PCI_SLOT(pdev->devfn);

        mb->board_enum = (int)golden_pci_board_id[tablecnt].driver_data;
        mb->pb_info = golden_pci_board_conf[mb->board_enum];     
		mb->boardFlag = mb->pb_info.boardFlag;
		
        mb->board_number = boardcnt - 1;        
    }   


    if (boardcnt == 0)
    {
        printk("GOLDEN Info : No Golden Series Multi-I/O Board Found !\n\n");
        status = -ENXIO;
        return status;
    }
	else
	{
		for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
		{
			mb = &golden_board_table[i];
			if (mb->board_enum > 0)
			{
        		printk("\n");			
				if ((mb->pb_info.numOfSerPort > 0) && (mb->pb_info.numOfParPort > 0))
				{
    				printk("GOLDEN Info : Found Golden %s Series Board (%dS%dP),\n", mb->pb_info.boardName, mb->pb_info.numOfSerPort, mb->pb_info.numOfParPort); 
				}
				else if ((mb->pb_info.numOfSerPort) > 0)
				{
					printk("GOLDEN Info : Found Golden %s Series Board (%dS),\n", mb->pb_info.boardName, mb->pb_info.numOfSerPort); 
				}
				else
				{
					printk("GOLDEN Info : Found Golden %s Series Board (%dP),\n", mb->pb_info.boardName, mb->pb_info.numOfParPort); 
				}
				
    			printk("              bus number:%d, device number:%d\n\n", mb->bus_number, mb->dev_number);
			}
		}
	}

    return status;
}


static int golden_get_pci_board_conf(void)
{
    struct golden_board *mb = NULL;
    struct pci_dev *pdev = NULL;
    int status = 0;
    int i;
    int j;
    
    for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
    {
        mb = &golden_board_table[i];
        
        if (mb->board_enum > 0)
        {
            pdev = mb->pdev;
            mb->ports = mb->pb_info.numOfSerPort + mb->pb_info.numOfParPort;
			mb->ser_port = mb->pb_info.numOfSerPort;
			mb->par_port = mb->pb_info.numOfParPort;
					
					
			snx_ser_port_total_cnt = snx_ser_port_total_cnt + mb->ser_port;
			snx_par_port_total_cnt = snx_par_port_total_cnt + mb->par_port;	
				
			if (snx_ser_port_total_cnt > GOLDEN_SER_TOTAL_MAX)
            {
                printk("GOLDEN Error: Too much serial port, maximum %d ports can be supported !\n\n", GOLDEN_SER_TOTAL_MAX);
                status = -EIO;
                return status;
            }	
				
			if (snx_par_port_total_cnt > GOLDEN_PAR_SUPPORT_MAX)
            {
                printk("GOLDEN Error: Too much parallel port, maximum %d ports can be supported !\n\n", GOLDEN_PAR_SUPPORT_MAX);
                status = -EIO;
                return status;
            }
            			
					
            for (j = 0; j < GOLDEN_PCICFG_BAR_TOTAL; j++) 
            {					
				mb->bar_addr[j] = pci_resource_start(pdev, j);	
			}
			
            mb->irq = mb->pdev->irq;    
			
            if (mb->irq <= 0)
            {
                printk("GOLDEN Error: Golden Board %s Series (bus:%d device:%d), in configuartion space, irq isn't valid !\n\n",
                                mb->pb_info.boardName,
                                mb->bus_number,
                                mb->dev_number
                                );

                status = -EIO;
                return status;
            }
        }
    }

    return status;
}


static int golden_assign_resource(void)
{
	struct golden_board *mb = NULL;
	struct golden_ser_port *sp = NULL;
	struct golden_par_port *pp = NULL;
    int status = 0;
	int i;
	int j;
	int k;
	int ser_n;
	
	memset(golden_ser_table, 0, (GOLDEN_SER_TOTAL_MAX + 1) * sizeof(struct golden_ser_port));
	memset(golden_par_table, 0, (GOLDEN_PAR_TOTAL_MAX) * sizeof(struct golden_par_port));
	
	for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
	{
		mb = &golden_board_table[i];
		
		if (mb->board_enum > 0)
		{
			if (mb->ser_port > 0)
			{
				mb->vector_mask = 0;
			
				// assign serial port resource
				ser_n = mb->ser_port_index = (mb->board_number * GOLDEN_SER_ONBOARD_MAX);
				sp = &golden_ser_table[ser_n];
			
				if (sp == NULL)
				{
					status = -ENXIO;
					printk("GOLDEN Error: Serial port table address error !\n");
					return status;
				}				
			
				for (j = 0; j < mb->ser_port; j++, ser_n++, sp++)
				{				
					sp->port.ChipFlag = mb->pb_info.port[j].ChipFlag;
					sp->port.iobase = mb->bar_addr[mb->pb_info.port[j].bar1] + mb->pb_info.port[j].offset1;
				
					if ((mb->boardFlag & BOARDFLAG_REMAP) == BOARDFLAG_REMAP)
					{
						sp->port.vector = 0;
					}
					else
					{
						sp->port.vector = mb->bar_addr[0] + SUN1889_INTRSERVREG;
					}
				
					mb->vector_mask  |= (1 << j);  
				}
			}
			
			
			// assign parallel port resource
			if (mb->par_port > 0)
			{
				k = 0;
				
				for (j = 0; j < GOLDEN_PAR_TOTAL_MAX; j++)
				{			
					if ((k + 1) > mb->par_port)
					{
						break;
					}
				
					if (j >= GOLDEN_PAR_TOTAL_MAX)
					{
						status = -EACCES;
						printk("GOLDEN Error: Too much parallel port, maximum %d ports can be supported !\n\n", GOLDEN_PAR_TOTAL_MAX);
						return status;
					}	
											
					pp = &golden_par_table[j];
			
					if (pp == NULL)
					{
						status = -ENXIO;
						printk("GOLDEN Error: Parallel port table address error !\n");
						return status;
					}				
			
					if (pp->ChipFlag != SUNNONE_HWID)
					{
						continue;
					}			
					else
					{	
						pp->ChipFlag = mb->pb_info.port[k + mb->ser_port].ChipFlag;		
						
						pp->base = mb->bar_addr[mb->pb_info.port[k + mb->ser_port].bar1] + mb->pb_info.port[k + mb->ser_port].offset1;
						pp->base_hi = mb->bar_addr[mb->pb_info.port[k + mb->ser_port].bar2] + mb->pb_info.port[k + mb->ser_port].offset2;
						
						pp->bus_number = mb->bus_number;
						pp->dev_number = mb->dev_number;
						pp->board_enum = mb->board_enum;
						k++;
					}
				}
			}	
		}
	}	
	
	return status;
}


static int golden_ser_port_table_init(void)
{
	struct golden_board *mb = NULL;
	struct golden_ser_port *sp = NULL;
    int status = 0;
	int i;
	int j;
	int n;

	for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
	{
		mb = &golden_board_table[i];
		
		if (mb == NULL)
		{
			status = -ENXIO;
			printk("GOLDEN Error: Board table pointer error !\n");
			return status;
		}	
				
		if ((mb->board_enum > 0) && (mb->ser_port > 0))
		{
			n = mb->ser_port_index = (mb->board_number * GOLDEN_SER_ONBOARD_MAX);
			sp = &golden_ser_table[n];

			if (sp == NULL)
			{
				status = -ENXIO;
				printk("GOLDEN Error: Serial port table pointer error !\n");
				return status;
			}
			
			for (j = 0; j < mb->ser_port; j++, n++, sp++) 
			{
				sp->port.board_enum			= mb->board_enum;
				sp->port.bus_number			= mb->bus_number;
				sp->port.dev_number			= mb->dev_number;
				sp->port.baud_base			= 921600;		
				sp->port.pb_info			= mb->pb_info;

				sp->port.irq 				= mb->irq;											
				sp->port.line 				= n;		
				sp->port.uartclk  			= sp->port.baud_base * 16;
				sp->port.iotype   			= SNX_UPIO_PORT;	
				sp->port.flags 				= ASYNC_SHARE_IRQ;	
				
				sp->port.ldisc_stop_rx		= 0;
				spin_lock_init(&sp->port.lock);
				
				if (sp->port.ChipFlag == SUN1889_HWID) 
				{
					sp->port.snx_type		= SNX_SER_PORT_SUN1889;
					sp->port.type			= PORT_SER_16650V2;
					sp->port.fifosize		= SUN1889_FIFOSIZE_SET;
					sp->port.rx_trigger		= SUN1889_TRIGGER_LEVEL_SET;
				}
				else if (sp->port.ChipFlag == SUN1699_HWID) 
				{
					sp->port.snx_type		= SNX_SER_PORT_SUN1699;
					sp->port.type			= PORT_SER_16650V2;
					sp->port.fifosize		= SUN1699_FIFOSIZE_SET;
					sp->port.rx_trigger		= SUN1699_TRIGGER_LEVEL_SET;
				}
				else if (sp->port.ChipFlag == SUN1999_HWID) 
				{
					sp->port.snx_type		= SNX_SER_PORT_SUN1999;
					sp->port.type			= PORT_SER_16750;
					sp->port.fifosize		= SUN1999_FIFOSIZE_SET;
					sp->port.rx_trigger		= SUN1999_TRIGGER_LEVEL_SET;					
				}				
				else
				{
					sp->port.snx_type		= SNX_SER_PORT_UNKNOWN;
					sp->port.type			= PORT_SER_16450;
					sp->port.fifosize		= DEFAULT_FIFOSIZE;
					sp->port.rx_trigger		= DEFAULT_TRIGGER_LEVEL;
				}
				
				
				if ((mb->pb_info.boardFlag & BOARDFLAG_REMAP) == BOARDFLAG_REMAP)
				{
					sp->port.vector_mask 	= 0;
					sp->port.PortFlag		= PORTFLAG_REMAP;
				}
				else
				{
					sp->port.vector_mask 	= mb->vector_mask;
					sp->port.PortFlag		= PORTFLAG_NONE;
				}
				
				sp->port.setserialFlag		= SNX_SER_BAUD_NOTSETSER;
			}		
			
			mb->SerIsrFunc = golden_ser_interrupt;
		}
		else
		{
			mb->SerIsrFunc = NULL;
		}		
	}
	
	
	// release io resource
	for (i = 0; i < GOLDEN_SER_TOTAL_MAX; i++)
	{
		sp = &golden_ser_table[i];
		
		if (sp->port.iobase > 0)
		{
			release_region(sp->port.iobase, GOLDEN_SER_ADDRESS_LENGTH);
		}
	}	
	
    return status;
}	


static int golden_par_port_table_init(void)
{
	struct golden_board *mb = NULL;
	struct golden_par_port *pp = NULL;
    int status = 0;
	int i;
	int j;
	int k;
	
	for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
	{
		mb = &golden_board_table[i];
		
		if (mb == NULL)
		{
			status = -ENXIO;
			printk("GOLDEN Error: Board table pointer error !\n");
			return status;
		}		
		
		if ((mb->board_enum > 0) && (mb->par_port > 0))
		{
			k = 0;
			for (j = 0; j < GOLDEN_PAR_TOTAL_MAX; j++)
			{
				pp = &golden_par_table[j];

				if (pp == NULL)
				{
					status = -ENXIO;
					printk("GOLDEN Error: Parallel port table pointer error !\n");
					return status;
				}				
			
				if ((k + 1) > mb->par_port)
				{
					break;
				}
			
			
				if ((pp->bus_number == mb->bus_number) && (pp->dev_number == mb->dev_number) && (pp->board_enum	= mb->board_enum))
				{
					pp->pb_info				= mb->pb_info;
					//pp->irq 				= mb->irq;											
					pp->irq					= PARPORT_IRQ_NONE;
					pp->portnum				= j;
					
					if (pp->ChipFlag == SUN1888_HWID) 
					{
						pp->snx_type		= SNX_PAR_PORT_SUN1888;
					}
					else if (pp->ChipFlag == SUN1689_HWID) 
					{
						pp->snx_type		= SNX_PAR_PORT_SUN1689;
					}
					else if (pp->ChipFlag == SUN1999_HWID) 
					{
						pp->snx_type		= SNX_PAR_PORT_SUN1999;
					}				
					else
					{
						pp->snx_type		= SNX_PAR_PORT_UNKNOWN;
					}
				
				
					if ((mb->pb_info.boardFlag & BOARDFLAG_REMAP) == BOARDFLAG_REMAP)
					{
						pp->PortFlag		= PORTFLAG_REMAP;
					}
					else
					{
						pp->PortFlag		= PORTFLAG_NONE;
					}	
					
					mb->ParIsrFunc = NULL;	
					k++;			
				}		
			}
		}
	}	
	
	
	// release io resource
	for (i = 0; i < GOLDEN_PAR_TOTAL_MAX; i++)
	{
		pp = &golden_par_table[i];
		
		if (pp->base > 0)
		{
			release_region(pp->base, GOLDEN_PAR_ADDRESS_LENGTH);

			
			release_region(pp->base, GOLDEN_PAR_STD_ADDR_LENGTH);
			release_region(pp->base + GOLDEN_PAR_STD_ADDR_LENGTH, GOLDEN_PAR_ETD_ADDR_LENGTH);

			if (pp->base_hi > 0)
			{
				release_region(pp->base_hi, GOLDEN_PAR_ADDRESS_LENGTH);	

				release_region(pp->base_hi, GOLDEN_PAR_STD_ADDR_LENGTH);
				release_region(pp->base_hi + GOLDEN_PAR_STD_ADDR_LENGTH, GOLDEN_PAR_ETD_ADDR_LENGTH);
			}
		}
	}

	return status;
}


int golden_register_irq(void)
{
	struct golden_board *mb = NULL;
	int status = 0;
	int i;
	
	for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
	{
		mb = &golden_board_table[i];
		
		if (mb == NULL)
		{
			status = -ENXIO;
			printk("GOLDEN Error: Board table pointer error !\n");
			return status;
		}			

		if (mb->board_enum > 0)
		{
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18))
			status = request_irq(mb->irq, golden_interrupt, IRQF_SHARED, "snx_golden", mb);
#else
            status = request_irq(mb->irq, golden_interrupt, SA_SHIRQ, "snx_golden", mb);
#endif	
		
			if (status) 
			{
	 			printk("GOLDEN Error: Golden Series Multi-I/O %s Board(bus:%d device:%d), request\n", mb->pb_info.boardName, mb->bus_number, mb->dev_number);	 			
	 			printk("              IRQ %d fail, IRQ %d may be conflit with another device.\n", mb->irq, mb->irq);
				return status;
			}		
		
		}
	}
	
	return status;
}


void golden_release_irq(void)
{
	struct golden_board *mb = NULL;
    int i;

    for (i = 0; i < GOLDEN_PCI_BOARDS_MAX; i++)
    {
        mb = &golden_board_table[i];
        
        if (mb->board_enum > 0)
        {
            free_irq(mb->irq, mb);
        }
    }
}


static struct snx_ser_driver    golden_ser_reg = 
{
    .dev_name       = "ttySNX",
    .major          = GOLDEN_TTY_MAJOR,
    .minor          = 0,
    .nr             = (GOLDEN_SER_TOTAL_MAX + 1),

};


static int __init snx_golden_init(void)
{
	int status = 0;

	printk("\n\n");
	printk("====================  GOLDEN Series Driver Module Install  =====================\n");
	printk("\n");
	printk("GOLDEN Info : Loading Golden Series Multi-I/O Board Driver Module\n");
	printk("                                                       -- Date :    %s\n",DRIVER_DATE);
	printk("                                                       -- Version : %s\n\n",DRIVER_VERSION);
	
	snx_ser_port_total_cnt = snx_par_port_total_cnt = 0;

	status = golden_pci_board_probe();
	if (status != 0)
	{
		goto step1_fail;
	}


	status = golden_get_pci_board_conf();
	if (status != 0)
	{
		goto step1_fail;
	}
	
	
	status = golden_assign_resource();
	if (status != 0)
	{
		goto step1_fail;
	}
	
	
	status = golden_ser_port_table_init();
	if (status != 0)
	{
		goto step1_fail;
	}
	
	
	status = golden_par_port_table_init();
	if (status != 0)
	{
		goto step1_fail;
	}


	status = golden_register_irq();
	if (status != 0)
	{
		goto step1_fail;
	}
	
	
	status = golden_ser_register_driver(&golden_ser_reg);
	if (status != 0)
	{
		goto step2_fail;
	}	


   	status = golden_ser_register_ports(&golden_ser_reg);
    if (status != 0)
    {
        goto step3_fail;
    }


	if (snx_par_port_total_cnt > 0)
	{
		status = golden_par_parport_init();
   		if (status != 0)
   		{
  			goto step4_fail;
   		}


		status = golden_par_ppdev_init();
   		if (status != 0)
   		{
   			goto step5_fail;
   		}	
   		
   		
		status = golden_par_lp_init();
   		if (status != 0)
   		{
   			goto step6_fail;
   		}	   		
	}
	
	printk("================================================================================\n");
	return status;


	if (snx_par_port_total_cnt > 0)
	{
step6_fail:

		golden_par_ppdev_exit();	
			
			
step5_fail:

		golden_par_parport_exit();	


step4_fail:

		golden_ser_unregister_ports(&golden_ser_reg);
	}

step3_fail:

	golden_ser_unregister_driver(&golden_ser_reg);
	

step2_fail:
	
	golden_release_irq();
	
	
step1_fail:

	printk("GOLDEN Error: Couldn't Loading Golden Series Multi-I/O Board Driver Module\n"); 
	printk("              correctly, please reboot system and try again, if error still\n");
	printk("              happen, contact support.\n\n");	
	printk("================================================================================\n");
	return status;

}


static void __exit snx_golden_exit(void)
{
	printk("\n\n");
	printk("===================  GOLDEN Series Driver Module Uninstall  ====================\n");
	printk("\n");

	if (snx_par_port_total_cnt > 0)
	{
		golden_par_lp_exit();
		
		golden_par_ppdev_exit();
	
		golden_par_parport_exit();
	}
	
	golden_ser_unregister_ports(&golden_ser_reg);
	
	golden_ser_unregister_driver(&golden_ser_reg);
	
	golden_release_irq();
	
	printk("GOLDEN Info : Unload Golden Series Multi-I/O Board Driver Module Done.\n");
	printk("================================================================================\n");
}

module_init(snx_golden_init);
module_exit(snx_golden_exit);

