/*
 *                                                                                      	
 *                      Port Information Dump Program                     
 *                                                                                      
 *				Copyright 2008 SUNIX GROUP all right reserved
 *                                                                                        
 *                                                              Version: 1.0.2.0           	
 *                                                              Date: 2008/09/05        
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <asm/ioctls.h>
#include <errno.h>


#define GOLDEN_SER_TOTAL_MAX	32	
#define GOLDEN_PAR_TOTAL_MAX	4		
#define MATRIX_SER_TOTAL_MAX	32	


#define GOLDEN  0x900
#define GOLDEN_SER_DUMP_PORT_INFO  (GOLDEN + 50)
#define GOLDEN_SER_DUMP_PORT_PERF  (GOLDEN + 51)
#define GOLDEN_SER_DUMP_DRIVER_VER (GOLDEN + 52)
#define GOLDEN_PAR_DUMP_PORT_INFO  (GOLDEN + 53)


#define MATRIX  0x700
#define MATRIX_SER_DUMP_PORT_INFO  (MATRIX + 50)
#define MATRIX_SER_DUMP_PORT_PERF  (MATRIX + 51)
#define MATRIX_SER_DUMP_DRIVER_VER (MATRIX + 52)



#define BOARDNAME_LENGTH  		15
#define DRIVERVERSION_LENGTH 	15
    
struct snx_ser_port_info 
{
 	char  			board_name_info[BOARDNAME_LENGTH];
 	unsigned int  	bus_number_info;
 	unsigned int  	dev_number_info;
 	unsigned int  	port_info;
 	unsigned int 	base_info;		
  	unsigned int  	irq_info;	
};


struct snx_par_port_info 
{
 	char  			board_name_info[BOARDNAME_LENGTH];
 	unsigned int  	bus_number_info;
 	unsigned int  	dev_number_info;
 	unsigned int  	port_info;
 	unsigned int  	base_info;
 	unsigned int	base_hi_info;		
 	unsigned int  	irq_info;	
};


struct snx_info
{
 	struct snx_ser_port_info 	golden_ser_port_info[GOLDEN_SER_TOTAL_MAX];
	struct snx_ser_port_info 	matrix_ser_port_info[MATRIX_SER_TOTAL_MAX];
	struct snx_par_port_info 	golden_par_port_info[MATRIX_SER_TOTAL_MAX];
	
	char 						golden_ver[DRIVERVERSION_LENGTH];
	char 						matrix_ver[DRIVERVERSION_LENGTH];
	
	int 						golden_ser_count;
	int 						matrix_ser_count;
	int 						golden_par_count;
	
	int 						golden_ser_status1;
	int							golden_ser_status2;
	int 						matrix_ser_status1;
	int							matrix_ser_status2;	
	int							golden_par_status1;
};


static int get_golden_ser_port_info(struct snx_info *infop)
{
	char golden_ser_name[14] = "/dev/ttySNX32";
	int g_ser_fd;
	
	g_ser_fd = open(golden_ser_name, O_RDWR | O_NOCTTY);

	if (g_ser_fd > 0)
	{
		infop->golden_ser_status1 = ioctl(g_ser_fd, GOLDEN_SER_DUMP_PORT_INFO, &infop->golden_ser_port_info);
		infop->golden_ser_status2 = ioctl(g_ser_fd, GOLDEN_SER_DUMP_DRIVER_VER, &infop->golden_ver);
		close(g_ser_fd);
		return 0;
	}
	
	return -ENXIO;
}


static int get_matrix_ser_port_info(struct snx_info *infop)
{
	char matrix_ser_name[14] = "/dev/ttySMX32";
    int m_ser_fd;
    
    m_ser_fd = open(matrix_ser_name, O_RDWR | O_NOCTTY);

    if (m_ser_fd > 0)
    {
        infop->matrix_ser_status1 = ioctl(m_ser_fd, MATRIX_SER_DUMP_PORT_INFO, &infop->matrix_ser_port_info);
        infop->matrix_ser_status2 = ioctl(m_ser_fd, MATRIX_SER_DUMP_DRIVER_VER, &infop->matrix_ver);
        close(m_ser_fd);
        return 0;
    }
    
    return -ENXIO;
}


static int get_golden_par_port_info(struct snx_info *infop)
{
	char golden_par_name[14] = "/dev/parport0";
	int g_par_fd;
	int i;
	
	for (i = 0; i < GOLDEN_PAR_TOTAL_MAX; i++)
	{
		golden_par_name[12] = '0' + i;
		
		g_par_fd = open(golden_par_name, O_WRONLY);
	
    	if (g_par_fd > 0)
    	{
        	infop->golden_par_status1 = ioctl(g_par_fd, GOLDEN_PAR_DUMP_PORT_INFO, &infop->golden_par_port_info[i]);
        	
        	close(g_par_fd);
    	}	
	}
	
	return 0;
}


static void print_info(struct snx_info *infop)
{
	int i;
	system("clear");
	
	for (i = 0; i < GOLDEN_SER_TOTAL_MAX; i++)
	{
		if (infop->golden_ser_port_info[i].base_info)
		{
			infop->golden_ser_count++;
		}
	}	

    for (i = 0; i < MATRIX_SER_TOTAL_MAX; i++)
    {
        if (infop->matrix_ser_port_info[i].base_info)
        {
            infop->matrix_ser_count++;
        }
    }  
    
    
    for (i = 0; i < GOLDEN_PAR_TOTAL_MAX; i++)
    {
        if (infop->golden_par_port_info[i].base_info)
        {
            infop->golden_par_count++;
        }
    }      
    
    
    
	if ((infop->golden_ser_count > 0) || (infop->matrix_ser_count > 0) || (infop->golden_par_count > 0))
	{
        if ((infop->golden_ser_count > 0) || (infop->golden_par_count > 0))
        {
		    printf("\n================ Found %2d GOLDEN port, list informations ====================\n", infop->golden_ser_count + infop->golden_par_count);
		    if (infop->golden_ser_status2 == 0)
			    printf("                                             GOLDEN driver ver -- %s\n\n", infop->golden_ver);
		    else
			    printf("\n");

		    for (i = 0; i < GOLDEN_SER_TOTAL_MAX; i++)
		    {
			    if (!(infop->golden_ser_port_info[i].base_info))
				    continue;
	
			    printf("ttySNX%d --\n", infop->golden_ser_port_info[i].port_info); 
			    printf("GOLDEN %s Series (bus:%d device:%2d) , base address = %4x,    irq = %2d\n\n",
					    infop->golden_ser_port_info[i].board_name_info,
 					    infop->golden_ser_port_info[i].bus_number_info,
 					    infop->golden_ser_port_info[i].dev_number_info,
 					    infop->golden_ser_port_info[i].base_info,
 					    infop->golden_ser_port_info[i].irq_info
					    );
		    }
		    
		    for (i = 0; i < GOLDEN_PAR_TOTAL_MAX; i++)
		    {
			    if (!(infop->golden_par_port_info[i].base_info))
				    continue;
	
			    printf("lp%d, parport%d --\n", infop->golden_par_port_info[i].port_info, infop->golden_par_port_info[i].port_info); 
			    printf("GOLDEN %s Series (bus:%d device:%2d) , base addr = %4x, extd addr = %4x\n\n",
					    infop->golden_par_port_info[i].board_name_info,
 					    infop->golden_par_port_info[i].bus_number_info,
 					    infop->golden_par_port_info[i].dev_number_info,
 					    infop->golden_par_port_info[i].base_info,
 					    infop->golden_par_port_info[i].base_hi_info,
 					    infop->golden_par_port_info[i].irq_info
					    );
		    }		    
        }

        if (infop->matrix_ser_count > 0)
        {
            printf("\n================ Found %2d MATRIX port, list informations ====================\n", infop->matrix_ser_count);
            if (infop->matrix_ser_status2 == 0)
                printf("                                             MATRIX driver ver -- %s\n\n", infop->matrix_ver);
            else
                printf("\n");

            for (i = 0; i < MATRIX_SER_TOTAL_MAX; i++)
            {
                if (!(infop->matrix_ser_port_info[i].base_info))
                    continue;
    
                printf("ttySMX%d --\n", infop->matrix_ser_port_info[i].port_info); 
                printf("MATRIX %s Series (bus:%d device:%2d) , base address = %4x,   irq = %2d\n\n",
                        infop->matrix_ser_port_info[i].board_name_info,
                        infop->matrix_ser_port_info[i].bus_number_info,
                        infop->matrix_ser_port_info[i].dev_number_info,
                        infop->matrix_ser_port_info[i].base_info,
                        infop->matrix_ser_port_info[i].irq_info
                        );
            }
        }
	}
	else
	{
        printf("\n============ No any port been found, type 'lsmod' to check driver ===========\n\n");
	}

	printf("=============================================================================\n");
}


int main(void)
{
	int status = 0;
	struct snx_info info;
	
	memset(&info, 0, (sizeof(struct snx_info)));
	
	status = get_golden_ser_port_info(&info);
	if (status < 0)
	{
	}
	
	status = get_matrix_ser_port_info(&info);
	if (status < 0)
	{
	}	
	
	status = get_golden_par_port_info(&info);
	if (status < 0)
	{
	}	
	
	print_info(&info);
	
 	return 0;
}

